/*

For basic usage instructions, check the /InputHandler interface definition.
You can treat /client objects as if they were of the type /InputHandler,
because this library makes /client implement the /InputHandler interface.

If you don't know how to use events, check out the Event library's demos.

Advanced users may want to override these procs:

* client.InputHandler_Initialize()
	Called in client/New() after ..() succeeds.
	Calls client.InputHandler_BindMacros().

* client.InputHandler_BindMacros()
	Called in client.InputHandler_Initialize() to create necessary macros.

* client.InputHandler_ClearDefaultMacros().
	Called in client.InputHandler_BindMacros() to clear default macros.

*/

#if !defined(TILE_WIDTH) || !defined(TILE_HEIGHT)
#error Define TILE_WIDTH and TILE_HEIGHT!
/*
You should have TILE_WIDTH and TILE_HEIGHT defined in your .dme above the BEGIN
INCLUDE line so that all included libraries see them.
They should be defined as the number of pixels matching your world.icon_size.
They are required to convert the position of the mouse from tile-step format
(x:step_x, y:step_y) to absolute pixel coordinates (px, py).
*/
#endif

/* InputHandler interface definition.
Anything that "looks like" an InputHandler can be used as an InputHandler.
It "looks like" an InputHandler if it has all of these events and procs defined.
*/
InputHandler
	var tmp/event
		/*
		Fires after the button [Button] is pressed.
		*/
		ButtonPressed // (InputHandler/InputHandler, Button)

		/*
		Fires after the button [Button] is released.
		*/
		ButtonReleased // (InputHandler/InputHandler, Button)

		/*
		Fires after a value of the axis [Axis] is changed.
		*/
		AxisChanged // (InputHandler/InputHandler, Axis, X, Y)

		/* Fires when the mouse wheel is scrolled.
		*/
		MouseScrolled // (InputHandler/InputHandler, DeltaX, DeltaY, Modifiers)

		/* Fires when the mouse is moved over the map.
		*/
		MouseMoved // (InputHandler/InputHandler, MoveX, MoveY)

	proc
		/* Returns TRUE while the button [Button] is pressed.
		*/
		IsButtonPressed(Button)

		/* Returns [list(x, y)] of the axis [Axis].
		[x] and [y] range from -1 (left/down) to 1 (right/up).
		*/
		GetAxisValues(Axis)

		/* Returns the position of the mouse in the screen, or null.
		Returns:
		* null if the mouse isn't over the map.
		* [list(x, y)] of the mouse's position relative to the
			bottom-left corner of the viewport (visible part of the map,
			excluding the extended HUD).
			[x] and [y] are in pixels.  Northeast is positive.
			list(1, 1) represents the bottom-left pixel.
		*/
		GetMouseScreenPosition()

		/* Returns the position of the mouse in the world, or null.
		Returns:
		* null if the mouse isn't over the map.
		* [list(x, y)] of the mouse's position relative to the bottom-left
			corner of the world map.
			[x] and [y] are in pixels. Northeast is positive.
			[list(1, 1)] is the bottom-left pixel of the turf [locate(1, 1, 1)].
		*/
		GetMouseMapPosition()

		/* Returns TRUE when the mouse is being tracked in the camera bounds.
		*/
		IsMouseInCamera()

		/*
		Returns a bit mask that contains held modifier keys.
		*/
		GetModifierMask(Modifiers)

/* Client implementation of InputHandler.
This lets you treat /client as an /InputHandler.
*/
client
	var tmp/event
		ButtonPressed
		ButtonReleased
		AxisChanged
		MouseScrolled
		MouseMoved

	proc
		ButtonPressed(InputHandler/InputHandler, Button)
			ButtonPressed && ButtonPressed.Fire(src, Button)

		ButtonReleased(InputHandler/InputHandler, Button)
			ButtonReleased && ButtonReleased.Fire(src, Button)

		AxisChanged(InputHandler/InputHandler, Axis, X, Y)
			AxisChanged && AxisChanged.Fire(src, Axis, X, Y)

		MouseScrolled(InputHandler/InputHandler, DeltaX, DeltaY, Modifiers)
			MouseScrolled && MouseScrolled.Fire(src, DeltaX, DeltaY, Modifiers)

		MouseMoved(InputHandler/InputHandler, MoveX, MoveY)
			MouseMoved && MouseMoved.Fire(src, MoveX, MoveY)

	proc
		IsButtonPressed(Button)
			return length(InputHandler_button_is_pressed) \
				&& InputHandler_button_is_pressed[Button] || 0

		GetAxisValues(Axis)
			return length(InputHandler_axis_values) > 0 \
				&& length(InputHandler_axis_values[Axis]) == 2 \
				&& InputHandler_axis_values[Axis] \
				|| list(0, 0)

		GetMouseScreenPosition()
			return InputHandler_mouse_screen_position.Copy()

		GetMouseMapPosition()
			var position[] = GetMouseScreenPosition()
			if(position)
				position[1] += bound_x - 1
				position[2] += bound_y - 1
				return position

		GetModifierMask()
			return \
				(IsButtonPressed(KeyButton.Control) \
					? ModifierKeys.Control : 0) \
				| (IsButtonPressed(KeyButton.Shift) \
					? ModifierKeys.Shift : 0) \
				| (IsButtonPressed(KeyButton.Alt) \
					? ModifierKeys.Alt : 0)

		IsMouseInCamera()
			return InputHandler_mouse_in_camera

/* Catches the mouse over the map if nothing else does.
*/
obj/InputHandler_mouse_catcher
	mouse_opacity = 2
	icon = null
	name = ""
	plane = -100
	layer = BACKGROUND_LAYER
	screen_loc = "SOUTHWEST to NORTHEAST"

/* Client InputHandler implementation details.
*/
client
	show_popup_menus = FALSE

	var tmp
		InputHandler_initialized
		InputHandler_button_is_pressed[]
		InputHandler_axis_values[]
		InputHandler_buttons_pressed[]
		InputHandler_mouse_screen_position[]
		InputHandler_mouse_in_camera

	var global
		obj/InputHandler_mouse_catcher/InputHandler_mouse_catcher
		regex/InputHandler_mouse_screen_loc_regex

	New()
		. = ..()
		if(.)
			InputHandler_Initialize()

	proc
		/* Start tracking macro input.
		*/
		InputHandler_Initialize()
			if(!InputHandler_mouse_screen_loc_regex)
				InputHandler_mouse_screen_loc_regex = regex("^(?:(?!\[0-9]+:)(\\w+):)?(\\d+)(?::(\\d+))?,(\\d+)(?::(\\d+))?$")

			if(!InputHandler_mouse_catcher)
				InputHandler_mouse_catcher = new /obj/InputHandler_mouse_catcher

			screen += InputHandler_mouse_catcher
			InputHandler_mouse_screen_position = list(1, 1)

			InputHandler_BindMacros()

			InputHandler_initialized = TRUE

		/* Creates macros to call verbs that update input states.
		*/
		InputHandler_BindMacros()
			InputHandler_ClearDefaultMacros()

			var id_prefix = "InputHandler\ref[src]"

			#if DM_VERSION >= 511
			// Track the Any key.
			winset(src, "[id_prefix]Any_key", "parent=macro; name=Any;\
				command=\".macro-button-down \[\[*]]\"")
			winset(src, "[id_prefix]Any_key_up", "parent=macro; name=Any+UP;\
				command=\".macro-button-up \[\[*]]\"")

			// Track gamepad axes.
			for(var/value in GamepadAxis.GetValues())
				winset(src, "[id_prefix][value]",
					"parent=macro; name=[value]; command=\".macro-axis \
					\\\"[value]\\\" \[\[x]] \[\[y]]\"")
			#else
			// Track all trackable keys.
			for(var/button in KeyButton.GetNames())
				var value = KeyButton.ToValue(button)
				winset(src, "[id_prefix][value]", "parent=macro; \
					name=[value]; command=\".macro-button-down \
					\\\"[value]\\\"")
				winset(src, "[id_prefix][value]_up", "parent=macro; \
					name=[value]+UP; command=\".macro-button-up \
					\\\"[value]\\\"")
			#endif

		/* Overwrite the default macros to do nothing.
		*/
		InputHandler_ClearDefaultMacros()
			var global/default_macros
			if(!default_macros)
				default_macros = list(
					"North+REP", "South+REP", "East+REP", "West+REP",
					"Northeast+REP", "Southeast+REP", "Northwest+REP",
					"Southwest+REP", "Center+REP",
					"F1", "F1+REP", // Options & Messages
					"F2", "F2+REP", // Screenshot Auto
					"Shift+F2", "Shift+F2+REP", // Screenshot
					// There doesn't seem to be a similar way to block Alt+F4.
					) - null
			for(var/macro in default_macros)
				winset(src, "macro_[replacetext(macro, "+", "_")]",
					"parent=macro; name=[macro]")

	verb
		InputHandler_MacroButtonDown(Button as text)
			set name = ".macro button down", instant = TRUE
			if(!InputHandler_buttons_pressed)
				InputHandler_buttons_pressed = new /list
				InputHandler_button_is_pressed = new /list
			if(InputHandler_buttons_pressed[Button]) return
			InputHandler_buttons_pressed[Button] = TRUE
			InputHandler_button_is_pressed[Button] = TRUE
			ButtonPressed(src, Button)

		InputHandler_MacroButtonUp(Button as text)
			set name = ".macro button up", instant = TRUE
			if(!length(InputHandler_buttons_pressed)) return
			if(!InputHandler_buttons_pressed[Button]) return
			InputHandler_buttons_pressed -= Button
			if(!length(InputHandler_buttons_pressed))
				InputHandler_buttons_pressed = null
				InputHandler_button_is_pressed = null
			else
				InputHandler_button_is_pressed[Button] = FALSE
			ButtonReleased(src, Button)

		InputHandler_MacroAxis(Axis as text, X as num, Y as num)
			set name = ".macro axis", instant = TRUE
			if(!InputHandler_axis_values)
				InputHandler_axis_values = new /list
			if(!InputHandler_axis_values[Axis])
				InputHandler_axis_values[Axis] = list(0, 0)
			var state[] = InputHandler_axis_values[Axis]
			if(state[1] != X || state[2] != Y)
				if(X == 0 && Y == 0)
					InputHandler_axis_values -= Axis
					if(!length(InputHandler_axis_values))
						InputHandler_axis_values = null
				else
					state[1] = X
					state[2] = Y
				AxisChanged(src, Axis, X, Y)

	MouseDown(object, location, control, params)
		InputHandler_initialized && InputHandler_MacroButtonDown(
			InputHandler_ExtractMouseButton(params2list(params)))
		..()

	MouseUp(object, location, control, params)
		InputHandler_initialized && InputHandler_MacroButtonUp(
			InputHandler_ExtractMouseButton(params2list(params)))
		..()

	MouseWheel(object, delta_x, delta_y, location, control, params)
		InputHandler_initialized && MouseScrolled(src, delta_x, delta_y,
			InputHandler_ExtractModifier(params2list(params)))
		..()

	MouseEntered(object, location, control, params)
		if(InputHandler_initialized)
			InputHandler_mouse_in_camera = isloc(location)
			InputHandler_CheckMouseMove(
				InputHandler_ExtractScreenPosition(params2list(params)))
		..()

	MouseExited(object, location, contrll, params)
		if(InputHandler_initialized)
			InputHandler_mouse_in_camera = FALSE
		..()

	MouseMove(object, location, control, params)
		InputHandler_initialized && InputHandler_CheckMouseMove(
			InputHandler_ExtractScreenPosition(params2list(params)))
		..()

	MouseDrag(src_object, over_object, src_location, over_location,
		src_control, over_control, params
		)
		InputHandler_initialized && InputHandler_CheckMouseMove(
			InputHandler_ExtractScreenPosition(params2list(params)))
		..()

	proc
		/* Compares ScreenPosition with the previous position, firing
		the MouseMoved event if they are different.
		*/
		InputHandler_CheckMouseMove(ScreenPosition[])
			if(length(ScreenPosition) < 2)
				return

			var
				old_x = InputHandler_mouse_screen_position[1]
				old_y = InputHandler_mouse_screen_position[2]
				new_x = ScreenPosition[1]
				new_y = ScreenPosition[2]

			if(new_x != old_x || new_y != old_y)
				var
					move_x = new_x - old_x
					move_y = new_y - old_y

				InputHandler_mouse_screen_position = ScreenPosition
				MouseMoved(src, move_x, move_y)

		/* Get the modifier flag from a mouse params list.
		*/
		InputHandler_ExtractModifier(Params[])
			return \
				(Params["ctrl"] ? ModifierKeys.Control : 0) | \
				(Params["shift"] ? ModifierKeys.Shift : 0) | \
				(Params["alt"] ? ModifierKeys.Alt : 0)

		/* Get the mouse button from a mouse params list.
		*/
		InputHandler_ExtractMouseButton(Params[])
			return \
				Params["left"] ? MouseButton.Left : \
				Params["right"] ? MouseButton.Right : \
				Params["middle"] ? MouseButton.Middle : null

		/* Get the screen position from a mouse params list.
		*/
		InputHandler_ExtractScreenPosition(Params[])
			if(InputHandler_mouse_screen_loc_regex.Find(Params["screen-loc"]))
				var data[] = InputHandler_mouse_screen_loc_regex.group
				return list(
					text2num(data[3]) + (text2num(data[2]) - 1) * TILE_WIDTH,
					text2num(data[5]) + (text2num(data[4]) - 1) * TILE_HEIGHT)
